import click
from cdislogging import get_logger
from jsonschema import ValidationError

from heal.vlmd.validate.validate import vlmd_validate

logging = get_logger("__name__")


@click.command()
@click.option(
    "--input_file",
    "input_file",
    required=True,
    help="name of file to validate",
    type=click.Path(writable=True),
)
def validate(input_file):
    """Validate VLMD input file"""

    logging.info(f"Validating VLMD file{input_file}")

    try:
        vlmd_validate(input_file)
        logging.info("Valid")
    except ValidationError as err:
        logging.error(f"Error in validating dictionary from {input_file}")
        logging.error(err.message)
    except Exception as e:
        logging.error(f"Validation error {str(e)}")
        logging.error("Invalid file")



from heal.vlmd.mappings.redcap_csv_headers import redcap_required_fields


def add_missing_type(prop_name: str, prop, schema: dict) -> dict:
    """
    Add types to properties.

    Args:
        prop_name (str): property name
        prop (str or dict): property
        schema (dict): schema

    Returns:
        schema (dict)

    """
    missing_values = ["", None]  # NOTE: include physical rep and logical for now
    if prop_name in schema.get("required", []):
        # if required value: MUST be NOT missing value and the property
        newprop = {"allOf": [prop, {"not": {"enum": missing_values}}]}
    else:
        # if not required value: MUST be property OR the specified missing value
        newprop = {"anyOf": [prop, {"enum": missing_values}]}
    return newprop


def add_types_to_props(schema: dict) -> dict:
    """
    Add missing types to the schema for validating csv style data.

    Args:
        schema (dict)

    Returns:
        schema (dict)
    """

    props_with_missing = {}
    for prop_name, prop in schema.get("items", {}).get("properties", {}).items():
        props_with_missing[prop_name] = add_missing_type(prop_name, prop, schema)

    patterns_with_missing = {}
    for pattern_name, prop in (
        schema.get("items", {}).get("patternProperties", {}).items()
    ):
        patterns_with_missing[pattern_name] = add_missing_type(
            pattern_name, prop, schema
        )

    schema = {"type": "array", "items": {}}
    schema["items"]["properties"] = props_with_missing
    schema["items"]["patternProperties"] = patterns_with_missing

    return schema


def remove_empty_props(props: dict) -> dict:
    """
    Remove items with empty values from a dict, ie,
    from a 'field' dict in a json dictionary.
    """
    if isinstance(props, dict):
        new_dict = {}
        for k, v in props.items():
            cleaned_value = remove_empty_props(v)
            if cleaned_value or cleaned_value == 0:
                new_dict[k] = cleaned_value
        return new_dict
    return props


def clean_json_fields(fields: list) -> list:
    """
    Remove any 'fields' with emtpy values.
    Can be used for json dictionaries that have been extracted from csv dictionaries.
    """
    cleaned_fields = []
    for field in fields:
        new_field = remove_empty_props(field)
        cleaned_fields.append(new_field)
    return cleaned_fields


def has_redcap_headers(column_names: list) -> bool:
    """
    Check if all required REDCap headers are present in the in column_names
    """
    return set(redcap_required_fields).issubset(column_names)



import click

from heal.cli import extract, validate


@click.group()
def main():
    """HEAL Command Line Interface"""
    pass


@click.group()
def vlmd():
    """Commands for VLMD"""
    pass


vlmd.add_command(extract.extract)
vlmd.add_command(validate.validate)



"""
Convert a REDCap data dictionary exported
in csv format to a heal-compliant json data dictionary.

STEPS
#1 fill section headers
#2 sort rows to allow proper row indexing
#3 make conditionals for each map function
"""

from os import PathLike

import numpy as np
import pandas as pd

import heal.vlmd.mappings.redcap_csv_headers as headers
from cdislogging import get_logger
from heal.vlmd.extract import utils
from heal.vlmd.extract.json_dict_conversion import convert_template_json
from heal.vlmd.mappings.redcap_field_mapping import type_mappings
from heal.vlmd.validate.utils import read_delim

logger = get_logger("redcap-conversion", log_level="debug")


def read_from_file(file_path):
    """
    Read csv from file.

    Returns dataframe.
    """
    source_dataframe = read_delim(file_path)
    return source_dataframe


def rename_and_fill(source_dataframe) -> list[dict]:
    """
    Rename columns to heal compliant, ffill

    Returns dict
    """
    source_dataframe = (
        source_dataframe.fillna("")
        .rename(columns=headers.mapping)
        .map(utils.strip_html)
    )

    # downfill section (if blank -- given we read in with petl, blanks are "" but ffill takes in np.nan)
    source_dataframe["section"] = (
        source_dataframe.replace({"": np.nan}).groupby("form")["section"].ffill()
    )
    # Recover any remaining np.nan (ie, not converted by 'ffill') back to "".
    # Include the 'infer_objects' method as described here
    # https://pandas.pydata.org/docs/whatsnew/v2.2.0.html#deprecated-automatic-downcasting
    source_dataframe = source_dataframe.fillna("").infer_objects(copy=False)
    return source_dataframe.to_dict(orient="records")


def _add_description(source_field: dict, target_field: dict) -> str:
    if source_field.get("label"):
        field_label = source_field["label"].strip()
    else:
        field_label = ""

    if source_field.get("section"):
        field_section = source_field["section"].strip() + ": "
    else:
        field_section = ""

    if target_field.get("description"):
        field_description = target_field["description"].strip()
    else:
        field_description = ""

    field_description = utils.strip_html(
        (field_section + field_label + field_description).strip()
    )
    if field_description:
        return field_description
    else:
        return "No field label for this variable"


def _add_title(source_field: dict, target_field: dict) -> str:
    target_title = target_field.get("title", "")

    if target_title:
        return target_title
    elif source_field.get("label"):
        return target_title + utils.strip_html(source_field["label"].strip())
    else:
        return "No field label for this variable"


def _add_section(source_field: dict):
    if source_field.get("form"):
        return source_field["form"]


def _add_metadata(source_field: dict, target_field: dict):
    target_field["description"] = _add_description(source_field, target_field)
    target_field["title"] = _add_title(source_field, target_field)
    target_field["section"] = _add_section(source_field)


def gather(source_fields: list[dict]) -> list[dict]:
    """Maps and translates fields from redcap to heal json"""

    source_data_fields = [
        field for field in source_fields if field["type"] in list(type_mappings)
    ]

    target_fields = []
    for source_field in source_data_fields:
        try:
            source_fieldtype = source_field["type"]
            target_field = type_mappings[source_fieldtype](source_field)
            # NOTE if one source_field generates more than 1 target field (ie checkbox) need to iterate through
            # if list (and hence not one to one mapping with source_field), assumes mandatory fields
            if isinstance(target_field, list):
                for _target_field in target_field:
                    assert "name" in _target_field and "type" in _target_field
                    _add_metadata(source_field, _target_field)
                    target_fields.append(_target_field)
            else:
                _add_metadata(source_field, target_field)
                target_field_with_name = {"name": source_field["name"]}
                target_field_with_name.update(target_field)
                target_fields.append(target_field_with_name)
        except Exception as err:
            logger.error(
                "REDCap conversion error for field '" f"{source_field['name']}" "'"
            )
            raise ValueError(str(err))

    return target_fields


def convert_redcap_csv(data_or_path, data_dictionary_props={}) -> dict[str, any]:
    """
    Takes in an exported Redcap Data Dictionary csv,
    and translates each field into a HEAL specified
    data dictionary based on field type (e.g., checkbox, radio, text and
    other conditional logic based on Redcap specifications.)


    > While there are a variety of options for Redcap exports (eg directly through
    the API or via an XML), the Redcap CSV provides an easy-to-edit format comfortable by
    technical and non-technical users.

    Parameters
    ----------
    data_or_path : str or path-like or a pandas DataFrame object.
        Data or path to data with the data being a tabular HEAL-specified data dictionary.
        This input can be any data object or path-like string excepted by a frictionless Resource object.
    data_dictionary_props : dict
        The HEAL-specified data dictionary properties.

    Returns
    -------
    dict
        A dictionary with two keys:
            - 'template_json': the HEAL-specified JSON object.
            - 'template_csv': the HEAL-specified tabular template.
    """

    if isinstance(data_or_path, (str, PathLike)):
        source_fields = read_from_file(data_or_path)
    elif not isinstance(data_or_path, pd.DataFrame):
        raise ValueError(
            "Input should be either dataframe or path to REDCap dictionary csv export"
        )
    else:
        source_fields = data_or_path
    source_fields = rename_and_fill(source_fields)
    target_fields = gather(source_fields)

    data_dictionary = data_dictionary_props.copy()
    data_dictionary["fields"] = target_fields

    package = convert_template_json(data_dictionary)
    return package



from os.path import isfile
from pathlib import Path
import re

from cdislogging import get_logger
from jsonschema import ValidationError

from heal.vlmd import ExtractionError, vlmd_validate
from heal.vlmd.config import (
    ALLOWED_FILE_TYPES,
    ALLOWED_INPUT_TYPES,
    ALLOWED_OUTPUT_TYPES,
    TOP_LEVEL_PROPS,
)
from heal.vlmd.extract.conversion import convert_to_vlmd
from heal.vlmd.file_utils import get_output_filepath, write_vlmd_dict
from heal.vlmd.validate.validate import file_type_to_fxn_map


logger = get_logger("extract", log_level="debug")


def set_title_if_missing(file_type: str, title: str, converted_dict: dict) -> dict:
    """
    Dictionary from json input should have an existing title or a user-supplied title.
    Dictionary from non-json input should set the title from
    standardsMappings[0].instrument.title or from the user supplied title.
    """

    existing_title = converted_dict.get("title")
    default_csv_title = TOP_LEVEL_PROPS.get("title")

    if file_type == "json":
        if existing_title is not None and existing_title != default_csv_title:
            logger.info(f"Converted dict already has a title {existing_title}")
            return converted_dict

        if title is not None:
            logger.debug(f"JSON dictionary setting user-defined title '{title}'")
            converted_dict["title"] = title
            return converted_dict

    else:
        instrument_title = None
        try:
            standards_mappings = converted_dict.get("standardsMappings")
            if standards_mappings and len(standards_mappings) >= 1:
                instrument_title = standards_mappings[0].get("instrument").get("title")
        except Exception as err:
            logger.warning("standardsMapping does not have 'instrument.title'")

        if title is None and instrument_title is None:
            message = "Title must be supplied when extracting from non-json to json"
            logger.error(message)
            raise ExtractionError(message)

        if instrument_title is not None:
            logger.debug(
                f"Converted CSV dictionary setting title to instrument title '{instrument_title}'"
            )
            converted_dict["title"] = instrument_title
            return converted_dict

        if title is not None and (
            existing_title is None or existing_title == default_csv_title
        ):
            logger.debug(
                f"Converted CSVdictionary setting user-defined title '{title}'"
            )
            converted_dict["title"] = title
            return converted_dict

    return converted_dict


def vlmd_extract(
    input_file, title=None, file_type="auto", output_dir=".", output_type="json"
) -> bool:
    """
    Extract a HEAL compliant csv and json format VLMD data dictionary
    from the input file.

    Args:
        input_file (str): the path of the input HEAL VLMD file to be extracted
            into HEAL-compliant VLMD file(s).
        title (str): the root level title of the dictionary (required if extracting from csv to json)
        file_type (str): the type of the input file that will be extracted into a
            HEAL-compliant VLMD file.
            Allowed values are "auto", “csv”, "json", "tsv", and "redcap"
            where "redcap" is a csv file of a REDCap dictionary export.
            Defaults to “auto”.
        output_dir (str): the directory of where the extracted VLMD file will
            be written. Defaults to “.”
        output_type (str): format of dictionary to write: "csv" or "json".
            The default is "json".

    Returns:
        True if the input is valid and is successfully converted and written.
        Raises ExtractionError in input VLMD is not valid or could not be converted.
    """

    logger.info(
        f"Extracting VLMD file '{input_file}' with input file_type '{file_type}'"
    )

    file_suffix = Path(input_file).suffix.replace(".", "")
    if file_suffix not in ALLOWED_INPUT_TYPES:
        message = f"Input file must be one of {ALLOWED_INPUT_TYPES}"
        logger.error(message)
        raise ExtractionError(message)

    if not isfile(input_file):
        message = f"Input file does not exist: {input_file}"
        logger.error(message)
        raise ExtractionError(message)

    if file_type not in ALLOWED_FILE_TYPES:
        message = f"File type must be one of {ALLOWED_FILE_TYPES}"
        logger.error(message)
        raise ExtractionError(message)

    if file_type == "auto":
        file_type = file_suffix

    if output_type not in ALLOWED_OUTPUT_TYPES:
        message = f"Unrecognized output_type '{output_type}' - should be in {ALLOWED_OUTPUT_TYPES}"
        logger.error(message)
        raise ExtractionError(message)

    if title is not None and re.match(r"^\s*$", title):
        message = f"Empty title is not allowed"
        logger.error(message)
        raise ExtractionError(message)

    logger.debug(f"File type is set to '{file_type}'")
    # validate
    try:
        # csv files are converted as part of validate
        converted_dictionary = vlmd_validate(
            input_file,
            file_type=file_type,
            output_type=output_type,
            return_converted_output=True,
        )
    except ValidationError as err:
        logger.error(f"Error in validating and extracting dictionary from {input_file}")
        logger.error(err.message)
        raise ExtractionError(str(err.message))
    except Exception as err:
        logger.error(f"Error in extracting dictionary from {input_file}")
        logger.error(err)
        raise ExtractionError(str(err))

    # input json file require explicit conversion and post validation steps
    if file_type == "json":
        file_convert_function = file_type_to_fxn_map.get(file_type)
        data_dictionary_props = {}
        try:
            logger.debug("Ready to convert json input to VLMD")
            data_dictionaries = convert_to_vlmd(
                input_filepath=input_file,
                input_type=file_convert_function,
                data_dictionary_props=data_dictionary_props,
            )
            if output_type == "json":
                converted_dictionary = data_dictionaries["template_json"]
                logger.debug(
                    f"Ready to validate converted dict with output type '{output_type}'"
                )
                is_valid = vlmd_validate(
                    converted_dictionary,
                    file_type=file_type,
                    output_type=output_type,
                    return_converted_output=False,
                )
                logger.debug(f"Converted dictionary is valid: {is_valid}")
            else:
                converted_dictionary = data_dictionaries["template_csv"]["fields"]
        except Exception as err:
            logger.error(f"Error in extracting JSON dictionary from {input_file}")
            logger.error(err)
            raise ExtractionError(str(err))

    if output_type == "json":
        converted_dictionary = set_title_if_missing(
            file_type=file_type, title=title, converted_dict=converted_dictionary
        )
        if converted_dictionary.get("title") is None:
            logger.error("JSON dictionary is missing 'title'")
            raise ExtractionError("JSON dictionary is missing 'title'")

    # write to file
    output_filepath = get_output_filepath(
        output_dir, input_file, output_type=output_type
    )
    logger.info(f"Writing converted dictionary to {output_filepath}")
    try:
        write_vlmd_dict(converted_dictionary, output_filepath, file_type=output_type)
    except Exception as err:
        logger.error("Error in writing converted dictionary")
        logger.error(err)
        raise ExtractionError("Error in writing converted dictionary")

    return True



"""
This module includes an external file retriever function intended to be called
by the external_files_download module in the Gen3-SDK.

The retriever function sends requests to the Syracuse QDR API for downloading studies or files.

The QDR documentation describes how to download studies
https://guides.dataverse.org/en/latest/api/dataaccess.html#basic-download-by-dataset

and how to download files
https://guides.dataverse.org/en/latest/api/dataaccess.html#basic-file-access

In order to get an API token from the WTS server, users should have already
sent a request to
<WTS-SERVER>/oauth2/authorization_url?idp=externaldata-keycloak
and logged in to QDR after the redirect.

The WTS-SERVER is a Gen3 commons that has been configured to return
tokens for the 'idp' specified in the external_file_metadata.
"""


from pathlib import Path
from typing import Dict, List
from heal.utils import unpackage_object, get_id, download_from_url, get_idp_access_token

from cdislogging import get_logger
from gen3.tools.download.drs_download import DownloadStatus, wts_get_token

logger = get_logger("__name__", log_level="debug")


def get_syracuse_qdr_files(
    wts_hostname: str, auth, file_metadata_list: List, download_path: str = "."
) -> Dict:
    """
    Retrieves external data from the Syracuse QDR.

    Args:
        wts_hostname (str): hostname for commons with wts
        auth (Gen3Auth): auth for commons with wts
        file_metadata_list (List of Dict): list of studies or files
        download_path (str): path to download files and unpack

    Returns:
        Dict of download status
    """
    if not Path(download_path).exists():
        logger.critical(f"Download path does not exist: {download_path}")
        return None

    completed = {}
    logger.debug(f"Input file metadata list={file_metadata_list}")

    for file_metadata in file_metadata_list:
        id = get_id(file_metadata)
        if id is None:
            logger.warning(
                f"Could not find 'study_id' or 'file_id' in metadata {file_metadata}"
            )
            continue
        logger.info(f"ID = {id}")
        completed[id] = DownloadStatus(filename=id, status="pending")

        download_url = get_download_url_for_qdr(file_metadata)
        if download_url is None:
            logger.critical(f"Could not get download_url for {id}")
            completed[id].status = "invalid url"
            continue

        idp_access_token = get_idp_access_token(wts_hostname, auth, file_metadata)
        request_headers = get_request_headers(idp_access_token)
        if "Authorization" not in request_headers:
            logger.critical("WARNING Request headers do not include a bearer token.")

        logger.debug(f"Request headers = {request_headers}")

        logger.debug(f"Ready to send request to download_url: GET {download_url}")
        downloaded_file = download_from_url(
            api_url=download_url,
            headers=request_headers,
            download_path=download_path,
        )
        if downloaded_file is None:
            completed[id].status = "failed"
            continue

        if downloaded_file.endswith("zip"):
            # unpack if download is zip file
            try:
                logger.debug(f"Ready to unpack {downloaded_file}.")
                unpackage_object(filepath=downloaded_file)
            except Exception as e:
                logger.critical(f"{id} had an issue while being unpackaged: {e}")
                completed[id].status = "failed"

            completed[id].status = "downloaded"
            # remove the zip file
            Path(downloaded_file).unlink()
        else:
            completed[id].status = "downloaded"

    if not completed:
        return None
    return completed


def get_download_url_for_qdr(file_metadata: Dict) -> str:
    """
    Get the download url for Syracuse QDR.

    Args:
        file_metadata (Dict)

    Returns:
        url, None if there are errors
    """
    base_url = "https://data.qdr.syr.edu/api/access"
    if "use_qdr_staging" in file_metadata and bool(file_metadata["use_qdr_staging"]):
        base_url = "https://data.stage.qdr.org/api/access"

    if "study_id" in file_metadata:
        url = f"{base_url}/dataset/:persistentId/?persistentId={file_metadata.get('study_id')}"
    elif "file_id" in file_metadata:
        url = f"{base_url}/datafile/{file_metadata.get('file_id')}"
    else:
        url = None

    return url


def get_request_headers(idp_access_token: str) -> Dict:
    """
    Generate the request headers.

    Args:
        idp_access_token (str): QDR access token is included in request header as bearer token
        file_metadata (Dict)

    Returns:
        Dictionary of request headers
    """
    headers = {}
    if idp_access_token:
        headers["Authorization"] = f"Bearer {idp_access_token}"

    return headers


def is_valid_qdr_file_metadata(file_metadata: Dict) -> bool:
    """
    Check that the file_metadata has the required keys:
    'study_id' or 'file_id'.

    Args:
        file_metadata (Dict)

    Returns:
        True if valid file_metadata object.
    """
    if not isinstance(file_metadata, dict):
        logger.critical(f"Invalid metadata - item is not a dict: {file_metadata}")
        return False
    if "study_id" not in file_metadata and "file_id" not in file_metadata:
        logger.critical(f"Invalid metadata - missing required QDR keys {file_metadata}")
        return False
    if "study_id" in file_metadata and "file_id" in file_metadata:
        logger.critical(
            f"Invalid metadata - item has both 'study_id' and 'file_id': {file_metadata}"
        )
        return False
    return True



import csv
import json
import os
from pathlib import Path

import pandas as pd
from cdislogging import get_logger

from heal.vlmd.config import OUTPUT_FILE_PREFIX

logger = get_logger("vlmd-file-utils", log_level="debug")


# file writing
def get_output_filepath(output_dir, infile_name, output_type="auto"):
    """
    Generate the output file path.
    Replace path to infile_name with output_dir.
    Replace suffix with 'json' if needed.

    Args:
      output_dir: output directory
      infile_name: the input file that dictionary is extracted from.
      output_type (str): the type of file to write to, currently csv and json.
        The default is "auto" which gets the type from the suffix of the input file name.

    Returns:
      string with output filepath:
          <output_dir>/<prefix>_<infile_name>.<suffix>
    """

    prefix = OUTPUT_FILE_PREFIX
    file_name = os.path.basename(infile_name)
    if output_type == "auto":
        output_filepath = f"{output_dir}/{prefix}_{file_name}"
    else:
        file_name = os.path.splitext(file_name)[0]
        output_filepath = f"{output_dir}/{prefix}_{file_name}.{output_type}"
    return output_filepath


def write_vlmd_dict(dictionary, output_filepath, file_type="auto"):
    """
    Write the json format dictionary to file.

    Args:
      dictionary (dict or array): data to write to file
      output_dir (path): path for output file
      file_type (str): type of file to write - "auto", "csv", "json".
          "auto" will get the type from the output_filepath suffix.
          The default is "auto".
    Returns:
      True for successful write of csv or json.
      None for unrecognized output type.
    """
    if not output_filepath:
        logger.error("Empty output_filepath in write_vlmd_dict")
        raise ValueError("Empty output_filepath in write_vlmd_dict")

    output_filepath = Path(output_filepath)

    if file_type == "auto":
        file_type = output_filepath.suffix.replace(".", "")

    logger.debug(f"Output dictionary type is {file_type}")
    dirname = os.path.dirname(output_filepath)
    if dirname != "":
        os.makedirs(dirname, exist_ok=True)

    if file_type == "json":
        if not isinstance(dictionary, dict):
            msg = "Json output type should have dict data"
            logger.error(msg)
            raise ValueError(msg)
        with open(output_filepath, "w") as f:
            json.dump(dictionary, f, indent=4)
        return True

    if file_type == "csv":
        if not isinstance(dictionary, list):
            msg = "CSV output type should have list data"
            logger.error(msg)
            raise ValueError(msg)
        quoting = csv.QUOTE_MINIMAL
        pd.DataFrame(dictionary).to_csv(output_filepath, quoting=quoting, index=False)
        return True

    logger.warning(f"Unknown file type {file_type}")
    return None



from functools import partial
from pathlib import Path

from cdislogging import get_logger

from heal.vlmd import mappings
from heal.vlmd.config import JSON_SCHEMA, TOP_LEVEL_PROPS
from heal.vlmd.extract.csv_dict_conversion import convert_datadict_csv
from heal.vlmd.extract.json_dict_conversion import convert_template_json
from heal.vlmd.extract.redcap_csv_dict_conversion import convert_redcap_csv
from heal.vlmd.utils import clean_json_fields

logger = get_logger("vlmd-conversion", log_level="debug")

choice_fxn = {
    "csv-data-dict": partial(
        convert_datadict_csv,
        rename_map=mappings.rename_map,
        recode_map=mappings.recode_map,
    ),
    "json-template": convert_template_json,
    "redcap-csv-dict": convert_redcap_csv,
}

ext_map = {
    ".csv": "csv-data-dict",
    ".json": "json-template",
}


def _detect_input_type(filepath, ext_to_input_type=ext_map):
    ext = filepath.suffix
    input_type = ext_to_input_type.get(ext, None)
    return input_type


def convert_to_vlmd(
    input_filepath,
    input_type=None,
    data_dictionary_props=None,
) -> dict:
    """
    Converts a data dictionary to HEAL compliant json or csv format.

    Args
        input_filepath (str): Path to input file. Currently converts data dictionaries in csv, json, and tsv.
        input_type (str): The input type. See keys of 'choice_fxn' dict for options, currently:
            csv-data-dict, json-template.
        data_dictionary_props (dict):
            The other data-dictionary level properties. By default,
            will give the data_dictionary `title` property as the file name stem.

    Returns
        Dictionary with:
         1. csvtemplated array of fields.
         2. jsontemplated data dictionary object as specified by an originally drafted design doc.
            That is, a dictionary with title:<title>,description:<description>,data_dictionary:<fields>
            where data dictionary is an array of fields as specified by the JSON schema.

    """

    input_filepath = Path(input_filepath)

    input_type = input_type or _detect_input_type(input_filepath)
    logger.debug(f"Converting file '{input_filepath}' of input_type '{input_type}'")
    if input_type not in choice_fxn.keys():
        logger.error(f"Unexpected input type {input_type}")
        raise ValueError(
            f"Unexpected input_type '{input_type}', not in {choice_fxn.keys()}"
        )

    # get data dictionary package based on the input type
    data_dictionary_props = data_dictionary_props or {}
    data_dictionary_package = choice_fxn[input_type](
        input_filepath, data_dictionary_props
    )

    # For now we return the csv and json in one package.
    # If any multiple data dictionaries are needed then implement the methods in
    # https://github.com/HEAL/healdata-utils/blob/5080227454d8e731d46a51aa6933c93523eb3b9a/src/healdata_utils/conversion.py#L196
    package = data_dictionary_package

    # add schema version
    for field in package["template_csv"]["fields"]:
        field.update({"schemaVersion": JSON_SCHEMA["version"], **field})

    # remove empty json fields, add schema version (in TOP_LEVEL_PROPS)
    package["template_json"]["fields"] = clean_json_fields(
        package["template_json"]["fields"]
    )
    package["template_json"] = {**TOP_LEVEL_PROPS, **dict(package["template_json"])}

    return package



condensed_rename_map = {
    "type": [
        "data-type",
        "data-types",
        "field-type",
        "field-types",
        "types",
        "variable-types",
        "variable-type",
    ],
    "name": [
        "variable-name",
        "variable-names",
        "column-name",
        "column-names",
        "field-name",
        "field-names",
    ],
    "title": [
        "variable-title",
        "variable-titles",
        "field-title",
        "field-titles",
        "field-label",
        "field-labels",
        "variable-label",
        "variable-labels",
    ],
    "description": [
        "definition",
        "definitions",
    ],
    "enumLabels": [
        "pv-description",
        "permissible-value-description",
        "variable-value-labels",
        "value-labels",
    ],
    "constraints.enum": [
        "admissible-values",
        "permissible-values",
        "possible-values",
        "enum",
        "enumerated-values",
        "valid-values",
    ],
}

rename_map = {
    source: target
    for target, source_list in condensed_rename_map.items()
    for source in source_list
}



import os
from os.path import isfile
from pathlib import Path

import jsonschema
from cdislogging import get_logger
from jsonschema import ValidationError

from heal.vlmd.config import (
    ALLOWED_FILE_TYPES,
    ALLOWED_INPUT_TYPES,
    ALLOWED_OUTPUT_TYPES,
    ALLOWED_SCHEMA_TYPES,
)
from heal.vlmd.extract.conversion import convert_to_vlmd
from heal.vlmd.utils import add_types_to_props
from heal.vlmd.validate.utils import get_schema, read_data_from_json_file, read_delim

logger = get_logger("vlmd-validate-extract", log_level="debug")


class ExtractionError(Exception):
    pass


file_type_to_fxn_map = {
    "csv": "csv-data-dict",
    "json": "json-template",
    "tsv": "csv-data-dict",
    "redcap": "redcap-csv-dict",
}


def vlmd_validate(
    input_file: str,
    file_type="auto",
    schema_type="auto",
    output_type="json",
    return_converted_output=False,
):
    """
    Validates the input file against a VLMD schema.

    Args:

        input_file (str): the path of the input HEAL VLMD file.
        file_type (str): the type of input file.
            Allowed values for now are “csv”, “tsv”, “json”, "redcap", and “auto”.
            Defaults to “auto” which will use the suffix of the input file.
            Using "auto" or "csv" will auto-detect if the file is a REDCap dictionary export.
        schema_type (str): the type of the schema to be validated against.
            Allowed values for now are “csv”, “tsv”, “json” and “auto”.
            Defaults to “auto” which will use the suffix of the input file.
        output_type (str): format of the converted dictionary: "csv" or "json.
            The default is "json".
        return_converted_output (bool): set to True to get converted output, else
            get a boolean for valid/invalid input.

    Returns:
        True if input is valid and return_converted_output=False.
        Returns a dictionary if input is valid and return_converted_output=True,
            where dictionary is converted if csv and dictionary is raw input if json.
        Raises ValidationError if the input VLMD is not valid.
        Raises ValueError for unallowed input file types or unallowed schema types.
        Raises SchemaError if the schema is invalid.
        Raises ExctractionError if the input cannot be converted to VLMD dictionary.
    """
    logger.debug("In vlmd validate")

    if isinstance(input_file, (str, os.PathLike)):
        logger.info(
            f"Validating VLMD file '{input_file}' against schema type '{schema_type}'"
        )
        if not isfile(input_file):
            message = f"Input file does not exist: {input_file}"
            logger.error(message)
            raise IOError(message)
        logger.debug(f"File_type = {file_type}")
        file_suffix = Path(input_file).suffix.replace(".", "")
        data = None
    else:
        logger.info(f"Validating VLMD json data against schema type '{schema_type}'")
        logger.debug(f"File_type = {file_type}")
        file_suffix = "json"
        data = input_file

    if file_type not in ALLOWED_FILE_TYPES:
        message = f"File type must be one of {ALLOWED_FILE_TYPES}"
        logger.error(message)
        raise ValueError(message)

    if file_suffix not in ALLOWED_INPUT_TYPES:
        message = f"Input file must be one of {ALLOWED_INPUT_TYPES}"
        logger.error(message)
        raise ValueError(message)

    if schema_type not in ALLOWED_SCHEMA_TYPES:
        message = f"Schema type must be in {ALLOWED_SCHEMA_TYPES}"
        logger.error(message)
        raise ValueError(message)

    schema = get_schema(input_file, schema_type)
    if schema is None:
        message = f"Could not get schema for type = {schema_type}"
        logger.error(message)
        raise ValueError(message)

    output_type = output_type if output_type else "json"
    if output_type not in ALLOWED_OUTPUT_TYPES:
        message = f"Unrecognized output_type '{output_type}' - should be in {ALLOWED_OUTPUT_TYPES}"
        logger.error(message)
        raise ValueError(message)

    if file_suffix in ["csv", "tsv"]:
        schema = add_types_to_props(schema)
    logger.debug("Checking schema")
    jsonschema.validators.Draft7Validator.check_schema(schema)

    # read the input file
    if file_suffix in ["csv", "tsv"]:
        logger.debug("Getting csv data from file")
        data = read_delim(input_file).to_dict(orient="records")
        if len(data) == 0:
            message = "Could not read csv data from input"
            logger.error(message)
            raise ValidationError(message)
    elif file_suffix == "json" and data is None:
        logger.debug("Getting json data from file")
        data = read_data_from_json_file(input_file)

    # if input is json then try a validation and return input
    if file_suffix == "json":
        logger.debug("Validating json data")
        try:
            jsonschema.validate(instance=data, schema=schema)
        except jsonschema.ValidationError as err:
            logger.error("Error in validating json input")
            raise err
        logger.debug("JSON input is valid")
        if return_converted_output:
            return data
        else:
            return True

    # else non-json data should first be converted and then validated
    if file_type == "auto":
        logger.debug(f"Using file_suffix for file_type {file_suffix}")
        file_convert_function = file_type_to_fxn_map.get(file_suffix)
    else:
        file_convert_function = file_type_to_fxn_map.get(file_type)
    if not file_convert_function:
        message = f"Could not get conversion function from file_suffix '{file_suffix}'"
        logger.error(message)
        raise ExtractionError(message)
    data_dictionaries = {}
    logger.debug(
        f"Verifying vlmd can be converted using file conversion function '{file_convert_function}'"
    )
    data_dictionary_props = {}
    try:
        data_dictionaries = convert_to_vlmd(
            input_filepath=input_file,
            input_type=file_convert_function,
            data_dictionary_props=data_dictionary_props,
        )
    except Exception as e:
        logger.error(f"Error in converting dictionary from {input_file}")
        logger.error(e)
        raise ExtractionError(str(e))
    if output_type == "json":
        converted_dictionary = data_dictionaries["template_json"]
    elif output_type == "csv":
        converted_dictionary = data_dictionaries["template_csv"]["fields"]

    # get a new schema if the output_type is different than input file_type.
    if output_type != file_suffix and not (
        file_suffix == "tsv" and output_type == "csv"
    ):
        logger.debug(
            f"Getting schema for '{output_type}' to validate converted dictionary"
        )
        schema = get_schema(converted_dictionary, schema_type=output_type)
        if output_type == "csv":
            schema = add_types_to_props(schema)
        if schema is None:
            message = f"Could not get schema for type = {schema_type}"
            logger.error(message)
            raise ValueError(message)

    try:
        jsonschema.validate(instance=converted_dictionary, schema=schema)
    except jsonschema.ValidationError as err:
        logger.error("Error in validating converted dictionary")
        raise err
    logger.debug("Converted dict is valid")

    if return_converted_output:
        return converted_dictionary
    else:
        return True



"""
holds the mappings for previous schema version
field names (headers, properties) and the current name
in addition to these field name values or value mappings.

"""
from heal.vlmd.config import JSON_SCHEMA

VERSION = ["0", "3"]

root_mappings = [
    {"target": "fields", "source": ["data_dictionary"]},
    {"target": "schemaVersion", "source": [None], "value": JSON_SCHEMA["version"]},
]


field_mappings = [
    {
        "target": "schemaVersion",
        "source": [None],
        # should be same as csv schema version
        "values": JSON_SCHEMA["version"],
    },
    {"target": "section", "source": ["module"]},
    {"target": "enumLabels", "source": ["encoding", "encodings"]},
    {"target": "enumOrdered", "source": ["ordered"]},
    {
        "target": "standardsMappings[0].instrument.source",
        "source": ["standardsMappings.type", "standardsMappings[0].source"],
    },
    {
        "target": "standardsMappings[0].instrument.title",
        "source": ["standardsMappings.label", "standardsMappings[0].title"],
    },
    {
        "target": "standardsMappings[0].item.source",
        "source": ["standardsMappings.source", "standardsMappings[0].source"],
    },
    {
        "target": "standardsMappings[0].item.id",
        "source": ["standardsMappings.id", "standardsMappings[0].id"],
    },
    {
        "target": "standardsMappings[0].item.url",
        "source": ["standardsMappings.url", "standardsMappings[0].url"],
    },
    {
        "target": None,
        "source": ["repo_link"],
    },
    {
        "target": None,
        "source": [
            "relatedConcepts.type",
            "relatedConcepts.label",
            "relatedConcepts.url",
            "relatedConcepts.source",
            "relatedConcepts.id",
            "relatedConcepts",
        ],
    },
    {
        "target": None,
        "source": [
            "univarStats.median",
            "univarStats.mean",
            "univarStats.std",
            "univarStats.min",
            "univarStats.max",
            "univarStats.mode",
            "univarStats.count",
            "univarStats.twentyFifthPercentile",
            "univarStats.seventyFifthPercentile",
            "univarStats.categoricalMarginals.name",
            "univarStats.categoricalMarginals.count",
        ],
    },
]


root_rename_map = {
    source_name: item["target"]
    for item in root_mappings
    for source_name in item["source"]
    if item["target"] and source_name
}
root_addmap = {"schemaVersion": JSON_SCHEMA["version"]}
fields_recode_map = {}
fields_rename_map = {
    source_name: item["target"]
    for item in field_mappings
    for source_name in item["source"]
    if item["target"] and source_name
}
fields_drop_list = [
    source_name
    for item in field_mappings
    for source_name in item["source"]
    if not item["target"] and source_name
]
fields_addmap = {"schemaVersion": JSON_SCHEMA["version"]}



mapping = {
    "Variable / Field Name": "name",
    "Form Name": "form",
    "Section Header": "section",
    "Field Type": "type",
    "Field Label": "label",
    "Choices, Calculations, OR Slider Labels": "choice_calc_lbls",
    "Field Note": "note",
    "Text Validation Type OR Show Slider Number": "text_valid_slider_num",
    "Text Validation Min": "text_valid_min",
    "Text Validation Max": "text_valid_max",
    "Identifier?": "identifier",
    "Branching Logic (Show field only if...)": "skip_logic",
    "Required Field?": "required",
    "Custom Alignment": "align",
    "Question Number (surveys only)": "question_num",
    "Matrix Group Name": "matrix_group",
}

CHOICES_LABEL_INPUT = "Choices, Calculations, OR Slider Labels"
CHOICES_FIELD_NAME = SLIDER_FIELD_NAME = CALC_FIELD_NAME = mapping[CHOICES_LABEL_INPUT]
TEXT_VALID_FIELD_NAME = mapping["Text Validation Type OR Show Slider Number"]

# This document lists 'Field Label' as required but Google AI excludes it
# https://ws.engr.illinois.edu/sitemanager/getfile.asp?id=1365
redcap_required_fields = [
    "Variable / Field Name",
    "Form Name",
    "Field Type",
    "Field Label",
]



"""Extract utilities/helper functions"""

import re
from collections.abc import MutableMapping

import jsonschema
import pandas as pd
from pandas.api.types import is_object_dtype


def _get_prop_names_to_rearrange(prop_names, schema):
    """
    Get data dictionary props to search for refactoring or embedding

    Supports patternProperties
    """

    annotation_names = ["title", "description", "name", "additionalDescription"]
    root_names = flatten_properties(schema["properties"]).keys()
    field_names = flatten_properties(
        schema["properties"]["fields"]["items"]["properties"]
    ).keys()
    props_to_rearrange = (
        set(root_names).intersection(field_names).difference(annotation_names)
    )

    prop_names_exp = "|".join("^" + name + "$" for name in props_to_rearrange)
    names_to_rearrange = [name for name in prop_names if re.match(prop_names_exp, name)]

    return list(names_to_rearrange)


def embed_data_dictionary_props(flat_fields, flat_root, schema):
    """
    Embed (flattened) root level props in each (flattened) field
    if field level prop is missing but the field level prop exists
    and is not an annotation property (title,description).

    Args

        flat_fields: array of dicts or pd.DataFrame (or something that can be converted into DataFrame)
        flat_root: dict or pd.Series or something that can be converted into Series
        schema: schema for determining what fields should be embedded (note, this is flattened in fxn)

    Returns

        pd.DataFrame with the flat fields with the embedded root properties
    """
    flat_fields = pd.DataFrame(flat_fields)
    prop_names = _get_prop_names_to_rearrange(list(flat_root.keys()), schema)
    flat_root = pd.Series(flat_root).loc[prop_names]  # take out annotation props
    if len(flat_root) > 0:
        for prop_name in prop_names:
            if prop_name in flat_root:
                if prop_name not in flat_fields:
                    flat_fields.insert(0, prop_name, flat_root[prop_name])
                else:
                    flat_fields[prop_name].fillna(flat_root[prop_name], inplace=True)

    return flat_fields


def refactor_field_props(flat_fields, schema):
    """
    Given a flattened array of dicts corresponding to the unflattened schema,
    move up (ie `refactor`) flattened properties that are both in the root
    (ie table level; level up from field records) and in the fields.
    """
    flat_fields_df = pd.DataFrame(flat_fields)
    props = set(_get_prop_names_to_rearrange(flat_fields_df.columns.tolist(), schema))
    flat_record = pd.Series(dtype="object")
    for name in props:
        in_df = name in flat_fields_df
        if in_df:
            is_one_unique = None
            # need to handle if some values are pandas series
            if isinstance(flat_fields_df[name], pd.DataFrame):
                is_one_unique = (flat_fields_df[name].nunique() == 1).all()
            elif isinstance(flat_fields_df[name], pd.Series):
                if is_object_dtype(flat_fields_df[name]):
                    is_one_unique = len(flat_fields_df[name].map(str).unique()) == 1
                else:
                    is_one_unique = flat_fields_df[name].nunique() == 1
            if is_one_unique:
                flat_record[name] = flat_fields_df.pop(name).iloc[0]
                if isinstance(flat_record[name], pd.Series):
                    flat_record[name] = flat_record[name].to_list()

    return flat_record, flat_fields_df


# individual cell utilities
def strip_html(html_string):
    """Strip out html from string"""
    if html_string:
        return re.sub(r"<[^>]+>", "", html_string)
    else:
        return html_string


def parse_dictionary_str(string, item_sep, key_val_sep):
    """
    Parses a stringified dictionary into a dictionary
    based on item separator
    """
    if string != "" and string is not None:
        str_items = string.strip().split(item_sep)
        items = {}

        for str_item in str_items:
            if str_item:
                item = str_item.split(key_val_sep, 1)
                items[item[0].strip()] = item[1].strip()

        return items
    else:
        return string


def parse_list_str(string, item_sep):
    """Split a string into list using 'item_sep'"""
    if string != "" and string is not None:
        return string.strip().split(item_sep)
    else:
        return string


# dictionary utilities
def flatten_to_json_path(dictionary, schema, parent_key=False, sep="."):
    """
    Turn a nested dictionary into a flattened dictionary (but see schema param)

    :param dictionary: The dictionary to flatten
    :param schema: The schema to indicate which properties to flatten
        This includes dictionaries (properties) with child dictionaries (properties)
        and lists (items) with child dictionaries (properties)
    :param parent_key: The string to prepend to dictionary's keys
    :param sep: The string used to separate flattened keys
    :return: A flattened dictionary
    """
    # flatten if type array -> type object with properties
    # flatten if type object with properties
    items = []
    for key, value in dictionary.items():
        new_key = str(parent_key) + sep + key if parent_key else key
        try:
            prop = schema["properties"]
        except:
            message = f"Missing key: 'properties' in prop {prop}"
            raise KeyError(message)
        prop = schema["properties"].get(key, {})
        childprops = prop.get("properties")
        child_item_props = prop.get("items", {}).get("properties")

        if child_item_props:
            for i, _value in enumerate(value):
                # check that child prop has properties
                try:
                    prop["properties"]
                except:
                    message = f"Missing key 'properties' in child prop: {_value}"
                    raise KeyError(message)
                item = flatten_to_json_path(
                    dictionary=_value,
                    schema=prop,
                    parent_key=new_key,
                    sep=f"[{str(i)}]{sep}",
                )
                items.extend(item.items())
        elif childprops:
            item = flatten_to_json_path(
                dictionary=value, schema=prop, parent_key=new_key, sep=sep
            )
            items.extend(item.items())

        else:
            items.append((new_key, value))

    return dict(items)


def unflatten_from_json_path(field):
    """
    Converts a flattened dictionary with key names conforming to
    JSONpath notation to the nested dictionary format.
    """
    field_json = {}

    for prop_path, prop in field.items():
        prop_json = field_json

        nested_names = prop_path.split(".")
        nested_names_length = len(nested_names)
        for i, prop_name in enumerate(nested_names):
            if "[" in prop_name:
                g = re.match(r"(.+)\[(\d+)\]$", prop_name)
                if not g:
                    raise jsonschema.ValidationError(
                        f"Incorrect array indexing in name {prop_name}"
                    )
                try:
                    array_name = g[1]
                    array_index = int(g[2])
                except Exception as e:
                    raise jsonschema.ValidationError(str(e))
                if array_name not in prop_json:
                    prop_json[array_name] = [{} for _ in range(array_index + 1)]
                elif len(prop_json[array_name]) <= array_index:
                    prop_json[array_name].extend(
                        [
                            {}
                            for _ in range(array_index - len(prop_json[array_name]) + 1)
                        ]
                    )
                if i == nested_names_length - 1:
                    if isinstance(prop_json[array_name][array_index], dict):
                        prop_json[array_name][array_index].update({array_name: prop})
                    else:
                        prop_json[array_name][array_index] = {array_name: prop}
                else:
                    prop_json = prop_json[array_name][array_index]
            else:
                if i == nested_names_length - 1:
                    prop_json[prop_name] = prop
                else:
                    if prop_name not in prop_json:
                        prop_json[prop_name] = {}
                    prop_json = prop_json[prop_name]

    return field_json


# json to csv utils
def join_iter(iterable, sep_list="|"):
    """Joins on iterable"""
    return sep_list.join([str(p) for p in iterable])


def join_dict_items(dictionary: dict, sep_key_val="=", sep_items="|"):
    """Joins a mappable collection (ie dictionary) into a string
    representation with specified separators for the key and value
    in addition to items.

    All items are coerced to the string representation (eg if key or value
    is None, this will be coerced to "None")
    """
    dict_list = []
    for key, val in dictionary.items():
        key_str = str(key)
        val_str = str(val)
        dict_list.append(key_str + sep_key_val + val_str)
    return sep_items.join(dict_list)


# Working with schemas
def flatten_properties(properties, parent_key="", sep=".", item_sep=r"\[\d+\]"):
    """Flatten schema properties"""
    properties_flattened = {}
    for key, item in properties.items():
        # flattened keys
        if parent_key:
            flattened_key = parent_key + "." + key
        else:
            flattened_key = key

        if isinstance(item, MutableMapping):
            props = item.get("properties")
            items = item.get("items", {}).get("properties")
            if props:
                new_props = flatten_properties(props, parent_key=flattened_key)
                properties_flattened.update(new_props)

            elif items:
                new_props = flatten_properties(
                    items, parent_key=flattened_key + item_sep
                )
                properties_flattened.update(new_props)
            else:
                properties_flattened[flattened_key] = item

        else:
            properties_flattened[flattened_key] = item

    return properties_flattened


def find_prop_name(column_name, properties):
    """
    Given a dictionary of json schema object properties OR a list of property names, return the
    matching property name.

    This function is needed when a schema is flattened according to a json path
    and converted into a regular expression for list (array) indices.

    """
    prop_match = []
    for name in list(properties):
        if re.match("^" + name + "$", column_name):
            prop_match.append(name)

    if len(prop_match) == 1:
        return prop_match[0]
    elif len(prop_match) > 1:
        raise Exception(
            f"Multiple matching properties found for {column_name}. Can only have one match"
        )
    elif len(prop_match) == 0:
        return None
    else:
        raise Exception(f"Unknown error when matching properties against {column_name}")
